##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
'''
Helper functions to create HyperLogLog operation dictionary arguments for
the :meth:`aerospike.operate` and :meth:`aerospike.operate_ordered` methods of the aerospike client.
HyperLogLog bins and operations allow for your application to form fast, reasonable approximations
of members in the union or intersection between multiple HyperLogLog bins.
HyperLogLog’s estimates are a balance between complete accuracy and efficient savings
in space and speed in dealing with extremely large datasets.

    .. note:: HyperLogLog operations require server version >= 4.9.0

    .. seealso:: `HyperLogLog (Data Type) more info. <https://www.aerospike.com/docs/guide/hyperloglog.html>`_.

Example::

    import sys

    import aerospike
    from aerospike import exception as ex
    from aerospike_helpers.operations import hll_operations as hll_ops
    from aerospike_helpers.operations import operations


    TEST_NS = "test"
    TEST_SET = "demo"
    NUM_INDEX_BITS = 12
    NUM_MH_BITS = 24

    # Configure the client.
    config = {"hosts": [("127.0.0.1", 3000)]}

    # Create a client and connect it to the cluster.
    try:
        client = aerospike.client(config).connect()
    except ex.ClientError as e:
        print("Error: {0} [{1}]".format(e.msg, e.code))
        sys.exit(1)

    # Add HLL bins.
    customers = ["Amy", "Farnsworth", "Scruffy"]
    customer_record_keys = [
        (TEST_NS, TEST_SET, "Amy"),
        (TEST_NS, TEST_SET, "Farnsworth"),
        (TEST_NS, TEST_SET, "Scruffy"),
    ]
    items_viewed = [
        ("item%s" % str(i) for i in range(0, 500)),
        ("item%s" % str(i) for i in range(0, 750)),
        ("item%s" % str(i) for i in range(250, 1000)),
    ]

    for customer, key, items in zip(customers, customer_record_keys, items_viewed):
        ops = [
            operations.write("name", customer),
            hll_ops.hll_add("viewed", list(items), NUM_INDEX_BITS, NUM_MH_BITS),
        ]

        try:
            client.operate(key, ops)
        except ex.ClientError as e:
            print("Error: {0} [{1}]".format(e.msg, e.code))
            sys.exit(1)

    # Find out how many items viewed Amy, Farnsworth, and Scruffy have in common.
    Farnsworth_viewed = client.get(customer_record_keys[1])[2]["viewed"]
    Scruffy_viewed = client.get(customer_record_keys[2])[2]["viewed"]
    viewed = [Farnsworth_viewed, Scruffy_viewed]
    ops = [hll_ops.hll_get_intersect_count("viewed", viewed)]

    try:
        _, _, res = client.operate(customer_record_keys[0], ops)
    except ex.ClientError as e:
        print("Error: {0} [{1}]".format(e.msg, e.code))
        sys.exit(1)

    print(
        "Estimated items viewed intersection: %d."
        % res["viewed"]
    )
    print("Actual intersection: 250.\\n")

    # Find out how many unique products Amy, Farnsworth, and Scruffy have viewed.
    Farnsworth_viewed = client.get(customer_record_keys[1])[2]["viewed"]
    Scruffy_viewed = client.get(customer_record_keys[2])[2]["viewed"]
    viewed = [Farnsworth_viewed, Scruffy_viewed]
    ops = [hll_ops.hll_get_union_count("viewed", viewed)]

    try:
        _, _, res = client.operate(customer_record_keys[0], ops)
    except ex.ClientError as e:
        print("Error: {0} [{1}]".format(e.msg, e.code))
        sys.exit(1)

    print(
        "Estimated items viewed union: %d."
        % res["viewed"]
    )
    print("Actual union: 1000.\\n")

    # Find the similarity of Amy, Farnsworth, and Scruffy's product views.
    Farnsworth_viewed = client.get(customer_record_keys[1])[2]["viewed"]
    Scruffy_viewed = client.get(customer_record_keys[2])[2]["viewed"]
    viewed = [Farnsworth_viewed, Scruffy_viewed]
    ops = [hll_ops.hll_get_similarity("viewed", viewed)]

    try:
        _, _, res = client.operate(customer_record_keys[0], ops)
    except ex.ClientError as e:
        print("Error: {0} [{1}]".format(e.msg, e.code))
        sys.exit(1)

    print(
        "Estimated items viewed similarity: %f%%."
        % (res["viewed"] * 100)
    )
    print("Actual similarity: 25%.")

    """
    Expected output:
    Estimated items viewed intersection: 235.
    Actual intersection: 250.

    Estimated items viewed union: 922.
    Actual union: 1000.

    Estimated items viewed similarity: 25.488069%.
    Actual similarity: 25%.
    """

'''

import aerospike


OP_KEY = "op"
BIN_KEY = "bin"
HLL_POLICY_KEY = "hll_policy"
INDEX_BIT_COUNT_KEY = "index_bit_count"
MH_BIT_COUNT_KEY = "mh_bit_count"
VALUE_LIST_KEY = "value_list"


def hll_add(bin_name: str, values, index_bit_count=None, mh_bit_count=None, policy=None):
    """Creates a hll_add operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server will add the values to the hll bin.
    If the HLL bin does not exist, it will be created with index_bit_count and/or mh_bit_count if they have been supplied.

    Args:
        bin_name (str): The name of the bin to be operated on.
        values: The values to be added to the HLL set.
        index_bit_count: An optional number of index bits. Must be bewtween 4 and 16 inclusive.
        mh_bit_count: An optional number of min hash bits. Must be bewtween 4 and 58 inclusive.
        policy (dict): An optional dictionary of :ref:`HyperLogLog policies <aerospike_hll_policies>`.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_ADD,
        BIN_KEY: bin_name,
        VALUE_LIST_KEY: values,
        INDEX_BIT_COUNT_KEY: -1 if index_bit_count is None else index_bit_count,
        MH_BIT_COUNT_KEY: -1 if mh_bit_count is None else mh_bit_count
    }

    if policy:
        op_dict[HLL_POLICY_KEY] = policy

    return op_dict


def hll_describe(bin_name):
    """Creates a hll_describe operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server returns index and minhash bit counts used to create HLL bin in a list of integers. 
    The list size is 2.

    Args:
        bin_name (str): The name of the bin to be operated on.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_DESCRIBE,
        BIN_KEY: bin_name,
    }

    return op_dict


def hll_fold(bin_name: str, index_bit_count):
    """Creates a hll_fold operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Servers folds index_bit_count to the specified value.
    This can only be applied when minhash bit count on the HLL bin is 0.
    Server does not return a value.

    Args:
        bin_name (str): The name of the bin to be operated on.
        index_bit_count: number of index bits. Must be bewtween 4 and 16 inclusive.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_FOLD,
        BIN_KEY: bin_name,
        INDEX_BIT_COUNT_KEY: index_bit_count
    }

    return op_dict


def hll_get_count(bin_name):
    """Creates a hll_get_count operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server returns estimated count of elements in the HLL bin. 

    Args:
        bin_name (str): The name of the bin to be operated on.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_GET_COUNT,
        BIN_KEY: bin_name,
    }

    return op_dict


def hll_get_intersect_count(bin_name: str, hll_list):
    """Creates a hll_get_intersect_count operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server returns estimate of elements that would be contained by the intersection of these HLL objects.

    Args:
        bin_name (str): The name of the bin to be operated on.
        hll_list (list): The HLLs to be intersected.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_GET_INTERSECT_COUNT,
        BIN_KEY: bin_name,
        VALUE_LIST_KEY: hll_list
    }

    return op_dict


def hll_get_similarity(bin_name: str, hll_list):
    """Creates a hll_get_similarity operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server returns estimated similarity of the HLL objects.
    Server returns a float.

    Args:
        bin_name (str): The name of the bin to be operated on.
        hll_list (list): The HLLs used for similarity estimation.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_GET_SIMILARITY,
        BIN_KEY: bin_name,
        VALUE_LIST_KEY: hll_list
    }

    return op_dict

def hll_get_union(bin_name: str, hll_list):
    """Creates a hll_get_union operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server returns an HLL object that is the union of all specified HLL objects
    in hll_list with the HLL bin.

    Args:
        bin_name (str): The name of the bin to be operated on.
        hll_list (list): The HLLs to be unioned.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_GET_UNION,
        BIN_KEY: bin_name,
        VALUE_LIST_KEY: hll_list
    }

    return op_dict

def hll_get_union_count(bin_name: str, hll_list):
    """Creates a hll_get_union_count operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server returns the estimated count of elements that would be contained by the union of all specified HLL objects
    in the list with the HLL bin.

    Args:
        bin_name (str): The name of the bin to be operated on.
        hll_list (list): The HLLs to be unioned.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_GET_UNION_COUNT,
        BIN_KEY: bin_name,
        VALUE_LIST_KEY: hll_list
    }

    return op_dict


def hll_init(bin_name: str, index_bit_count=None, mh_bit_count=None, policy=None):
    """Creates a hll_init operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server creates a new HLL or resets an existing HLL.
    If index_bit_count and mh_bit_count are None, an existing HLL bin will be reset but retain its configuration.
    If 1 of index_bit_count or mh_bit_count are set,
    an existing HLL bin will set that config and retain its current value for the unset config.
    If the HLL bin does not exist, index_bit_count is required to create it, mh_bit_count is optional.
    Server does not return a value.

    Args:
        bin_name (str): The name of the bin to be operated on.
        index_bit_count: An optional number of index bits. Must be bewtween 4 and 16 inclusive.
        mh_bit_count: An optional number of min hash bits. Must be bewtween 4 and 58 inclusive.
        policy (dict): An optional dictionary of :ref:`HyperLogLog policies <aerospike_hll_policies>`.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_INIT,
        BIN_KEY: bin_name,
        INDEX_BIT_COUNT_KEY: -1 if index_bit_count is None else index_bit_count,
        MH_BIT_COUNT_KEY: -1 if mh_bit_count is None else mh_bit_count
    }

    if policy:
        op_dict[HLL_POLICY_KEY] = policy

    return op_dict


def hll_refresh_count(bin_name: str):
    """Creates a hll_refresh_count operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server updates the cached count if it is stale.
    Server returns the count. 

    Args:
        bin_name (str): The name of the bin to be operated on.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_REFRESH_COUNT,
        BIN_KEY: bin_name,
    }

    return op_dict

def hll_set_union(bin_name: str, hll_list, policy=None):
    """Creates a hll_set_union operation to be used with :meth:`aerospike.operate` or :meth:`aerospike.operate_ordered`.

    Server sets the union of all specified HLL objects with the HLL bin.
    Server returns nothing.

    Args:
        bin_name (str): The name of the bin to be operated on.
        hll_list (list): The HLLs who's union will be set.
        policy (dict): An optional dictionary of :ref:`HyperLogLog policies <aerospike_hll_policies>`.
    """
    op_dict = {
        OP_KEY: aerospike.OP_HLL_SET_UNION,
        BIN_KEY: bin_name,
        VALUE_LIST_KEY: hll_list
    }

    if policy:
        op_dict[HLL_POLICY_KEY] = policy

    return op_dict



##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
'''
HyperLogLog expressions contain expressions for performing HLL operations.
Most of these operations are equivalent to the :mod:`HyperLogLog API <aerospike_helpers.operations.hll_operations>`.

Example::

    import aerospike_helpers.expressions as exp
    # Get count from HLL bin "d".
    expr = exp.HLLGetCount(exp.HLLBin("d")).compile()
'''

#from __future__ import annotations
from itertools import chain
from typing import List, Optional, Tuple, Union, Dict, Any
import aerospike
from aerospike_helpers import cdt_ctx
from aerospike_helpers.expressions.resources import _BaseExpr
from aerospike_helpers.expressions.resources import ResultType
from aerospike_helpers.expressions.resources import _Keys
from aerospike_helpers.expressions.base import HLLBin

########################
# HLL Modify Expressions
########################

TypeBinName = Union[_BaseExpr, str]
TypeListValue = Union[_BaseExpr, List[Any]]
TypeValue = Union[_BaseExpr, Any]
TypePolicy = Union[Dict[str, Any], None]


class HLLInit(_BaseExpr):
    """
       Creates a new HLL or resets an existing HLL.
       If index_bit_count and mh_bit_count are None, an existing HLL bin will be reset but retain its configuration.
       If 1 of index_bit_count or mh_bit_count are set,
       an existing HLL bin will set that config and retain its current value for the unset config.
       If the HLL bin does not exist, index_bit_count is required to create it, mh_bit_count is optional.
    """
    _op = aerospike.OP_HLL_INIT

    def __init__(self, policy: 'TypePolicy', index_bit_count: Union[int, None], mh_bit_count: Union[int, None], bin: 'TypeBinName'):
        """ Args:
                policy (TypePolicy): An optional dictionary of :ref:`HyperLogLog policies <aerospike_hll_policies>`.
                index_bit_count (int): Number of index bits. Must be between 4 and 16 inclusive.
                mh_bit_count (int): Number of min hash bits. Must be between 4 and 51 inclusive.
                bin (TypeBinName): An :class:`~aerospike_helpers.expressions.base.HLLBin` expression.

            :return: Returns the resulting hll.

            Example::

                # Create an HLL with 12 index bits and 24 min hash bits.
                expr = exp.HLLInit(None, 12, 24, exp.HLLBin("my_hll"))
        """
        self._children = (
            -1 if index_bit_count is None else index_bit_count,
            -1 if mh_bit_count is None else mh_bit_count,
            policy['flags'] if policy is not None and 'flags' in policy else aerospike.HLL_WRITE_DEFAULT,
            bin if isinstance(bin, _BaseExpr) else HLLBin(bin)
        )


class HLLAdd(_BaseExpr):
    """Create an expression that performs an hll_add."""
    _op = aerospike.OP_HLL_ADD

    def __init__(self, policy: 'TypePolicy', list: 'TypeListValue', index_bit_count: Union[int, None], mh_bit_count: Union[int, None], bin: 'TypeBinName'):
        """ Args:
                policy (TypePolicy): An optional dictionary of :ref:`HyperLogLog policies <aerospike_hll_policies>`.
                list (TypeListValue): A list or list expression of elements to add to the HLL.
                index_bit_count (int): Number of index bits. Must be between 4 and 16 inclusive.
                mh_bit_count (int): Number of min hash bits. Must be between 4 and 51 inclusive.
                bin (TypeBinName): An :class:`~aerospike_helpers.expressions.base.HLLBin` expression.

            :return: Returns the resulting hll bin after adding elements from list.

            Example::

                # Let HLL bin "d" have the following elements, ['key1', 'key2', 'key3'], index_bits 8, mh_bits 8.
                # Add ['key4', 'key5', 'key6'] so that the returned value is ['key1', 'key2', 'key3', 'key4', 'key5', 'key6']
                expr = exp.HLLAdd(None, ['key4', 'key5', 'key6'], 8, 8, exp.HLLBin("d")).compile()
        """
        self._children = (
            list,
            -1 if index_bit_count is None else index_bit_count,
            -1 if mh_bit_count is None else mh_bit_count,
            policy['flags'] if policy is not None and 'flags' in policy else aerospike.HLL_WRITE_DEFAULT,
            bin if isinstance(bin, _BaseExpr) else HLLBin(bin)
        )


######################
# HLL Read Expressions
######################


class HLLGetCount(_BaseExpr):
    """Create an expression that performs an as_operations_hll_get_count."""
    _op = aerospike.OP_HLL_GET_COUNT

    def __init__(self, bin: 'TypeBinName'):
        """ Args:
               bin (TypeBinName): An :class:`~aerospike_helpers.expressions.base.HLLBin` expression.

            :return: Integer bin, the estimated number of unique elements in an HLL.

            Example::

                # Get count from HLL bin "d".
                expr = exp.HLLGetCount(exp.HLLBin("d")).compile()
        """
        self._children = (
            bin if isinstance(bin, _BaseExpr) else HLLBin(bin),
        )


class HLLGetUnion(_BaseExpr):
    """Create an expression that performs an hll_get_union."""
    _op = aerospike.OP_HLL_GET_UNION

    def __init__(self, values: 'TypeValue', bin: 'TypeBinName'):
        """ Args:
                values (TypeValue): A single HLL or list of HLLs, values or expressions, to union with bin.
                bin (TypeBinName): An :class:`~aerospike_helpers.expressions.base.HLLBin` expression.

            :return: HLL bin representing the set union.

            Example::

                # Let HLLBin "d" contain keys ['key%s' % str(i) for i in range(10000)].
                # Let values be a list containing HLL objects retrieved from the aerospike database.
                # Find the union of HLL bin "d" and all HLLs in values.
                expr = exp.HLLGetUnion(values, exp.HLLBin("d")).compile()
        """
        self._children = (
            values,
            bin if isinstance(bin, _BaseExpr) else HLLBin(bin),
        )


class HLLGetUnionCount(_BaseExpr):
    """Create an expression that performs an as_operations_hll_get_union_count."""
    _op = aerospike.OP_HLL_GET_UNION_COUNT

    def __init__(self, values: 'TypeValue', bin: 'TypeBinName'):
        """ Args:
                values (TypeValue): A single HLL or list of HLLs, values or expressions, to union with bin.
                bin (TypeBinName): An :class:`~aerospike_helpers.expressions.base.HLLBin` expression.

            :return: Integer bin, estimated number of elements in the set union.

            Example::

                # Let HLLBin "d" contain keys ['key%s' % str(i) for i in range(10000)].
                # Let values be a list containing one HLL object with keys ['key%s' % str(i) for i in range(5000, 15000)].
                # Find the count of keys in the union of HLL bin "d" and all HLLs in values. (Should be around 15000)
                expr = exp.HLLGetUnionCount(values, exp.HLLBin("d")).compile()
        """
        self._children = (
            values,
            bin if isinstance(bin, _BaseExpr) else HLLBin(bin),
        )


class HLLGetIntersectCount(_BaseExpr):
    """Create an expression that performs an as_operations_hll_get_inersect_count."""
    _op = aerospike.OP_HLL_GET_INTERSECT_COUNT

    def __init__(self, values: 'TypeValue', bin: 'TypeBinName'):
        """ Args:
                values (TypeValue): A single HLL or list of HLLs, values or expressions, to intersect with bin.
                bin (TypeBinName): An :class:`~aerospike_helpers.expressions.base.HLLBin` expression.

            :return: Integer bin, estimated number of elements in the set intersection.

            Example::

                # Let HLLBin "d" contain keys ['key%s' % str(i) for i in range(10000)].
                # Let values be a list containing one HLL object with keys ['key%s' % str(i) for i in range(5000, 15000)].
                # Find the count of keys in the intersection of HLL bin "d" and all HLLs in values. (Should be around 5000)
                expr = exp.HLLGetIntersectCount(values, exp.HLLBin("d")).compile()
        """
        self._children = (
            values,
            bin if isinstance(bin, _BaseExpr) else HLLBin(bin),
        )


class HLLGetSimilarity(_BaseExpr):
    """Create an expression that performs an as_operations_hll_get_similarity."""
    _op = aerospike.OP_HLL_GET_SIMILARITY

    def __init__(self, values: 'TypeValue', bin: 'TypeBinName'):
        """ Args:
                values (TypeValue): A single HLL or list of HLLs, values or expressions, to calculate similarity with.
                bin (TypeBinName): An :class:`~aerospike_helpers.expressions.base.HLLBin` expression.

            :return: Float bin, stimated similarity between 0.0 and 1.0.

            Example::

                # Let HLLBin "d" contain keys ['key%s' % str(i) for i in range(10000)].
                # Let values be a list containing one HLL object with keys ['key%s' % str(i) for i in range(5000, 15000)].
                # Find the similarity the HLL in values to HLL bin "d". (Should be around 0.33)
                # Note that similarity is defined as intersect(A, B, ...) / union(A, B, ...).
                expr = exp.HLLGetSimilarity(values, exp.HLLBin("d")).compile()
        """
        self._children = (
            values,
            bin if isinstance(bin, _BaseExpr) else HLLBin(bin),
        )


class HLLDescribe(_BaseExpr):
    """Create an expression that performs an as_operations_hll_describe."""
    _op = aerospike.OP_HLL_DESCRIBE

    def __init__(self, bin: 'TypeBinName'):
        """ Args:
                bin (TypeBinName): An :class:`~aerospike_helpers.expressions.base.HLLBin` expression.

            :return: List bin, a list containing the index_bit_count and minhash_bit_count.

            Example::

                # Get description of HLL bin "d".
                expr = exp.HLLDescribe(exp.HLLBin("d")).compile()
        """
        self._children = (
            bin if isinstance(bin, _BaseExpr) else HLLBin(bin),
        )


class HLLMayContain(_BaseExpr):
    """ Create an expression that checks if the HLL bin contains any keys in
        list.
    """
    _op = aerospike.OP_HLL_MAY_CONTAIN

    def __init__(self, list: 'TypeListValue', bin: 'TypeBinName', ):
        """ Args:
                list (TypeListValue): A list expression of keys to check if the HLL may contain them.
                bin (TypeBinName): An :class:`~aerospike_helpers.expressions.base.HLLBin` expression.

            :return: 1 if bin contains any key in list, 0 otherwise.

            Example::

                # Check if HLL bin "d" contains any of the keys in `list`.
                expr = exp.HLLMayContain(["key1", "key2", "key3"], exp.HLLBin("d")).compile()
        """
        self._children = (
            list,
            bin if isinstance(bin, _BaseExpr) else HLLBin(bin),
        )



# -*- coding: utf-8 -*-
################################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

from __future__ import print_function
import os
import platform
import sys
from subprocess import Popen
from subprocess import call
from setuptools import setup, Extension
from distutils.command.build import build
from distutils.command.clean import clean
from multiprocessing import cpu_count
import time
import io

################################################################################
# ENVIRONMENT VARIABLES
################################################################################

os.putenv('ARCHFLAGS', '-arch x86_64')
os.environ['ARCHFLAGS'] = '-arch x86_64'
AEROSPIKE_C_VERSION = os.getenv('AEROSPIKE_C_VERSION')
BASEPATH = os.path.dirname(os.path.abspath(__file__))
AEROSPIKE_C_HOME = os.path.join(BASEPATH, 'aerospike-client-c')

AEROSPIKE_C_TARGET = None
PLATFORM = platform.platform(1)
LINUX = 'Linux' in PLATFORM
DARWIN = 'Darwin' in PLATFORM or 'macOS' in PLATFORM
CWD = os.path.abspath(os.path.dirname(__file__))
STATIC_SSL = os.getenv('STATIC_SSL')
SSL_LIB_PATH = os.getenv('SSL_LIB_PATH')
EVENT_LIB = os.getenv('EVENT_LIB')

################################################################################
# GENERIC BUILD SETTINGS
################################################################################

include_dirs = ['src/include'] + \
    [x for x in os.getenv('CPATH', '').split(':') if len(x) > 0] + \
    ['/usr/local/opt/openssl/include'] + \
    ['aerospike-client-c/modules/common/src/include']
extra_compile_args = [
    '-std=gnu99', '-g', '-Wall', '-fPIC', '-O1', '-DDEBUG',
    '-fno-common', '-fno-strict-aliasing', '-Wno-strict-prototypes',
    '-march=nocona',
    '-D_FILE_OFFSET_BITS=64', '-D_REENTRANT',
    '-DMARCH_x86_64',
    '-Wno-implicit-function-declaration'
]
extra_objects = []
extra_link_args = []
library_dirs = ['/usr/local/opt/openssl/lib', '/usr/local/lib']
libraries = [
    'ssl',
    'crypto',
    'pthread',
    'm',
    'z'
]
################################################################################
# STATIC SSL LINKING BUILD SETTINGS
################################################################################

if STATIC_SSL:
    extra_objects.extend(
        [SSL_LIB_PATH + 'libssl.a', SSL_LIB_PATH + 'libcrypto.a'])
    libraries.remove('ssl')
    libraries.remove('crypto')
    library_dirs.remove('/usr/local/opt/openssl/lib')

################################################################################
# PLATFORM SPECIFIC BUILD SETTINGS
################################################################################

if DARWIN:
    # ---------------------------------------------------------------------------
    # Mac Specific Compiler and Linker Settings
    # ---------------------------------------------------------------------------
    extra_compile_args = extra_compile_args + [
        '-D_DARWIN_UNLIMITED_SELECT'
    ]

    AEROSPIKE_C_TARGET = AEROSPIKE_C_HOME + '/target/Darwin-x86_64'

elif LINUX:
    # ---------------------------------------------------------------------------
    # Linux Specific Compiler and Linker Settings
    # ---------------------------------------------------------------------------
    extra_compile_args = extra_compile_args + [
        '-rdynamic', '-finline-functions'
    ]
    libraries = libraries + ['rt']
    AEROSPIKE_C_TARGET = AEROSPIKE_C_HOME + '/target/Linux-x86_64'
else:
    print("error: OS not supported:", PLATFORM, file=sys.stderr)
    sys.exit(8)

include_dirs = include_dirs + [
    '/usr/local/opt/openssl/include',
    AEROSPIKE_C_TARGET + '/include'
    ]
extra_objects = extra_objects + [
    AEROSPIKE_C_TARGET + '/lib/libaerospike.a'
]

os.putenv('CPATH', ':'.join(include_dirs))
os.environ['CPATH'] = ':'.join(include_dirs)

################################################################################
# SETUP
################################################################################

# Get the long description from the relevant file
with io.open(os.path.join(CWD, 'README.rst'), "r", encoding='utf-8') as f:
    long_description = f.read()

# Get the version from the relevant file
with io.open(os.path.join(CWD, 'VERSION'), "r", encoding='utf-8') as f:
    version = f.read()

BASEPATH = os.path.dirname(os.path.abspath(__file__))
CCLIENT_PATH = os.path.join(BASEPATH, 'aerospike-client-c')

# if EVENT_LIB is None or EVENT_LIB == "":
#     EVENT_LIB = "libevent"

if EVENT_LIB is not None:
    if EVENT_LIB == "libuv":
        extra_compile_args = extra_compile_args + ['-DAS_EVENT_LIB_DEFINED']
        library_dirs = library_dirs + ['/usr/local/lib/']
        libraries = libraries + ['uv']
    elif EVENT_LIB == "libevent":
        extra_compile_args = extra_compile_args + ['-DAS_EVENT_LIB_DEFINED']
        library_dirs = library_dirs + ['/usr/local/lib/']
        libraries = libraries + ['event_core', 'event_pthreads']
    elif EVENT_LIB == "libev":
        extra_compile_args = extra_compile_args + ['-DAS_EVENT_LIB_DEFINED']
        library_dirs = library_dirs + ['/usr/local/lib/']
        libraries = libraries + ['ev']
    else:
        print("Building aerospike with no-async support\n")

class CClientBuild(build):

    def run(self):
        if self.force == 1:
            # run original c-extension clean task
            # clean.run(self)
            cmd = [
                'make',
                'clean'
            ]
            def clean():
                call(cmd, cwd=CCLIENT_PATH)
            self.execute(clean, [], 'Clean core aerospike-client-c previous builds')

        os.putenv('LD_LIBRARY_PATH', ':'.join(library_dirs))
        os.environ['LD_LIBRARY_PATH'] = ':'.join(library_dirs)
        os.putenv('DYLD_LIBRARY_PATH', ':'.join(library_dirs))
        os.environ['DYLD_LIBRARY_PATH'] = ':'.join(library_dirs)
        # build core client
        cmd = [
            'make',
            'V=' + str(self.verbose),
        ]
        if EVENT_LIB is not None:
            cmd = [
                'make',
                'V=' + str(self.verbose),
                'EVENT_LIB='+EVENT_LIB] 

        def compile():
            print(cmd, library_dirs, libraries)
            call(cmd, cwd=CCLIENT_PATH)

        self.execute(compile, [], 'Compiling core aerospike-client-c')
        # run original c-extension build code
        build.run(self)


class CClientClean(clean):

    def run(self):
        # run original c-extension clean task
        # clean.run(self)
        cmd = [
            'make',
            'clean'
        ]

        def clean():
            call(cmd, cwd=CCLIENT_PATH)

        self.execute(clean, [], 'Clean core aerospike-client-c')


setup(
    name='aerospike',
    version=version.strip(),
    description='Aerospike Client Library for Python',
    long_description=long_description,
    author='Aerospike, Inc.',
    author_email='info@aerospike.com',
    url='http://aerospike.com',
    license='Apache Software License',
    keywords=['aerospike', 'nosql', 'database'],
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Database'
    ],

    # Package Data Files
    zip_safe=False,
    include_package_data=True,

    # Data files
    ext_modules=[
        Extension(
            # Extension Name
            'aerospike',

            # Source Files
            [
                'src/main/aerospike.c',
                'src/main/exception.c',
                'src/main/log.c',
                'src/main/client/type.c',
                'src/main/client/apply.c',
                'src/main/client/bit_operate.c',
                'src/main/client/cdt_list_operate.c',
                'src/main/client/cdt_map_operate.c',
                'src/main/client/hll_operate.c',
                'src/main/client/expression_operations.c',
                'src/main/client/cdt_operation_utils.c',
                'src/main/client/close.c',
                'src/main/client/connect.c',
                'src/main/client/exists.c',
                'src/main/client/exists_many.c',
                'src/main/client/get.c',
                'src/main/client/get_async.c',
                'src/main/client/put_async.c',
                'src/main/client/get_many.c',
                'src/main/client/batch_get_ops.c',
                'src/main/client/select_many.c',
                'src/main/client/info_single_node.c',
                'src/main/client/info_random_node.c',
                'src/main/client/info_node.c',
                'src/main/client/info.c',
                'src/main/client/put.c',
                'src/main/client/operate_list.c',
                'src/main/client/operate_map.c',
                'src/main/client/operate.c',
                'src/main/client/query.c',
                'src/main/client/remove.c',
                'src/main/client/scan.c',
                'src/main/client/select.c',
                'src/main/client/tls_info_host.c',
                'src/main/client/truncate.c',
                'src/main/client/admin.c',
                'src/main/client/udf.c',
                'src/main/client/sec_index.c',
                'src/main/serializer.c',
                'src/main/client/remove_bin.c',
                'src/main/client/get_key_digest.c',
                'src/main/query/type.c',
                'src/main/query/apply.c',
                'src/main/query/add_ops.c',
                'src/main/query/foreach.c',
                'src/main/query/predexp.c',
                'src/main/query/results.c',
                'src/main/query/select.c',
                'src/main/query/where.c',
                'src/main/query/execute_background.c',
                'src/main/scan/type.c',
                'src/main/scan/foreach.c',
                'src/main/scan/results.c',
                'src/main/scan/select.c',
                'src/main/scan/execute_background.c',
                'src/main/scan/apply.c',
                'src/main/scan/add_ops.c',
                'src/main/scan/paginate.c',
                'src/main/geospatial/type.c',
                'src/main/geospatial/wrap.c',
                'src/main/geospatial/unwrap.c',
                'src/main/geospatial/loads.c',
                'src/main/geospatial/dumps.c',
                'src/main/policy.c',
                'src/main/conversions.c',
                'src/main/convert_predexp.c',
                'src/main/convert_expressions.c',
                'src/main/policy_config.c',
                'src/main/calc_digest.c',
                'src/main/predicates.c',
                'src/main/tls_config.c',
                'src/main/global_hosts/type.c',
                'src/main/nullobject/type.c',
                'src/main/cdt_types/type.c',
                'src/main/key_ordered_dict/type.c',
                'src/main/client/set_xdr_filter.c',
                'src/main/client/get_nodes.c',
                'src/main/convert_partition_filter.c',
                'src/main/client/get_key_partition_id.c'
            ],

            # Compile
            include_dirs=include_dirs,
            extra_compile_args=extra_compile_args,

            # Link
            library_dirs=library_dirs,
            libraries=libraries,
            extra_objects=extra_objects,
            extra_link_args=extra_link_args,
        )
    ],
    packages=['aerospike_helpers', 'aerospike_helpers.operations',
              'aerospike_helpers.expressions', 'aerospike_helpers.awaitable'],

    cmdclass={
        'build': CClientBuild,
        'clean': CClientClean
    }
)



##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################
'''
Module with helper functions to do async get/put by
the :mod:`aerospike.Client.awaitable` methods for the aerospike.client class.
'''
import asyncio
import aerospike
from aerospike import exception as e

_get_results = {}
_get_cqd = 0
_put_results = {}
_put_cqd = 0

async def _io_get_put(client, op=0, key=None, record=None, meta=None, policy=None, serialize=None):
    global _get_cqd, _put_cqd
    index = None

    if key is not None:
        if len(key) >= 3:
            index = key[2]
        if index is None:
            if len(key) >= 4:
                index = key[3]

    if index is not None and isinstance(index, bytearray):
        index = bytes(index)

    #print(f"io:_io_get_put op:{op} key:{key} index:{index}")

    def put_async_callback(key_tuple, err, exce):
        global _put_cqd

        _put_cqd -= 1

        result = [key_tuple, err, exce]
        _put_results[index]['result'] = result

        #print(f"io:put_cb {_put_results[index]}")

        loop = _put_results[index]['loop']
        fut = _put_results[index]['fut']

        if result[1][0] != 0:
            #print(f"io:put_cb except:{result[2]}")
            loop.call_soon_threadsafe(fut.set_exception, result[2])
        else:
            loop.call_soon_threadsafe(fut.set_result, result[1][0])

    def get_async_callback(key_tuple, record_tuple, err, exce):
        global _get_cqd

        _get_cqd -= 1

        result = [key_tuple, record_tuple, err, exce]
        _get_results[index]['result'] = result

        #print(f"io:get_cb {_get_results[index]}")

        loop = _get_results[index]['loop']
        fut = _get_results[index]['fut']

        if result[2][0] != 0:
            #print(f"io:get_cb except:{result[3]}")
            loop.call_soon_threadsafe(fut.set_exception, result[3])
        else:
            loop.call_soon_threadsafe(fut.set_result, result[1])

    loop = asyncio.get_event_loop()
    future = loop.create_future()
    
    context = {'loop': None, 'fut': None, 'result': None}

    if index is not None:
        if op == 0:
            _put_results[index] = context
            _put_results[index]['loop'] = loop
            _put_results[index]['fut'] = future
            _put_cqd += 1
        elif op == 1:
            _get_results[index] = context
            _get_results[index]['loop'] = loop
            _get_results[index]['fut'] = future
            _get_cqd += 1

    try:
        if op == 0:
            client.put_async(put_async_callback, key, record, meta, policy, serialize)
        if op == 1:
            client.get_async(get_async_callback, key, policy)
    except Exception as e:
        #print(f"io:post_get key:{key} except:{e} cqd:{[_get_cqd if op else _put_cqd]}")
        raise e

    await future
    
    if index is not None:
        if op == 0:
            del _put_results[index]
        if op == 1:
            del _get_results[index]
    
    return future.result()

async def get(client, key=None, policy=None):
    return await _io_get_put(client, 1, key, None, None, policy)

async def put(client, key=None, record=None, meta=None, policy=None, serialize=None):
    return await _io_get_put(client, 0, key, record, meta, policy, serialize)



'''
Resources used by all expressions.
'''

#from __future__ import annotations
from itertools import chain
from typing import List, Optional, Tuple, Union, Dict, Any
import aerospike
from aerospike_helpers import cdt_ctx


class _Keys:
    VALUE_TYPE_KEY = "value_type"
    BIN_KEY = "bin"
    RETURN_TYPE_KEY = "return_type"
    CTX_KEY = "ctx"
    VALUE_KEY = "val"
    LIST_POLICY_KEY = "list_policy"
    MAP_POLICY_KEY = "map_policy"
    LIST_ORDER_KEY = "list_order"
    REGEX_OPTIONS_KEY = "regex_options"


class _ExprOp: # TODO replace this with an enum
    UNKNOWN = 0

    EQ = 1
    NE = 2
    GT = 3
    GE = 4
    LT = 5
    LE = 6
    CMP_REGEX = 7
    CMP_GEO = 8

    AND = 16
    OR = 17
    NOT = 18
    EXCLUSIVE = 19

    ADD = 20
    SUB = 21
    MUL = 22
    DIV = 23
    POW = 24
    LOG = 25
    MOD = 26
    ABS = 27
    FLOOR = 28
    CEIL = 29

    TO_INT = 30
    TO_FLOAT = 31

    INT_AND = 32
    INT_OR = 33
    INT_XOR = 34
    INT_NOT = 35
    INT_LSHIFT = 36
    INT_RSHIFT = 37
    INT_ARSHIFT = 38
    INT_COUNT = 39
    INT_LSCAN = 40
    INT_RSCAN = 41

    MIN = 50
    MAX = 51

    META_DIGEST_MOD = 64
    META_DEVICE_SIZE = 65
    META_LAST_UPDATE_TIME = 66
    META_VOID_TIME = 67
    META_TTL = 68
    META_SET_NAME = 69
    META_KEY_EXISTS = 70
    META_SINCE_UPDATE_TIME = 71
    META_IS_TOMBSTONE = 72

    REC_KEY = 80
    BIN = 81
    BIN_TYPE = 82
    BIN_EXISTS = 83

    COND = 123
    VAR = 124
    LET = 125
    DEF = 126

    # virtual ops

    _AS_EXP_CODE_CALL_VOP_START = 139
    _AS_EXP_CODE_CDT_LIST_CRMOD = 140
    _AS_EXP_CODE_CDT_LIST_MOD = 141
    _AS_EXP_CODE_CDT_MAP_CRMOD = 142
    _AS_EXP_CODE_CDT_MAP_CR = 143
    _AS_EXP_CODE_CDT_MAP_MOD = 144

    _AS_EXP_CODE_END_OF_VA_ARGS = 150

    _TRUE = 151
    _FALSE = 152

    _AS_EXP_BIT_FLAGS = 153

    VAL = 200


class ResultType:
    """
    Flags used to indicate expression value_type.
    """
    BOOLEAN = 1
    INTEGER = 2
    STRING = 3
    LIST = 4
    MAP = 5
    BLOB = 6
    FLOAT = 7
    GEOJSON = 8
    HLL = 9


class _AtomExpr:
    def _op(self):
        raise NotImplementedError

    def compile(self):
        raise NotImplementedError


TypeResultType = Optional[int]
TypeFixedEle = Union[int, float, str, bytes, dict]
TypeFixed = Optional[Dict[str, TypeFixedEle]]
TypeCompiledOp = Tuple[int, TypeResultType, TypeFixed, int]
TypeExpression = List[TypeCompiledOp]

TypeChild = Union[int, float, str, bytes, _AtomExpr]
TypeChildren = Tuple[TypeChild, ...]

TypeAny = Union[_AtomExpr, Any]

class _BaseExpr(_AtomExpr):
    _op = 0
    # type: int
    _rt = None
    # type: 'TypeResultType'
    _fixed = None
    # type: 'TypeFixed'
    _children = ()
    # type: 'TypeChildren'

    def _get_op(self) -> TypeCompiledOp:
        return (self._op, self._rt, self._fixed, len(self._children))

    def _vop(self, v) -> TypeCompiledOp:
        return (
            _ExprOp.VAL,
            None,
            {_Keys.VALUE_KEY: v},
            0,
        )

    def compile(self) -> TypeExpression:
        expression = [self._get_op()]
        # type: 'TypeExpression'
        work = chain(self._children)

        while True:
            try:
                item = next(work)
            except StopIteration:
                break

            if isinstance(item, _BaseExpr):
                expression.append(item._get_op())
                work = chain(item._children, work)
            else:
                # Should be a str, bin, int, float, etc.
                expression.append(self._vop(item))

        return expression

    def _overload_op_unary(self, op_type: int):
        if self._op == op_type:
            l = self._children
        else:
            l = (self,)

        r = [] # No right operand.
        return _create_operator_expression(l, r, op_type)

    def _overload_op(self, right: 'TypeAny', op_type: int):
        if self._op == op_type:
            l = self._children
        else:
            l = (self,)

        if isinstance(right, _BaseExpr) and right._op == op_type:
            r = right._children
        else:
            r = (right,)

        return _create_operator_expression(l, r, op_type)

    def _overload_op_va_args(self, right: 'TypeAny', op_type: int):
        expr_end = _BaseExpr()
        expr_end._op = _ExprOp._AS_EXP_CODE_END_OF_VA_ARGS

        if self._op == op_type:
            # Last element of an expression with var args'
            # children will always be _AS_EXP_CODE_END_OF_VA_ARGS.
            l = self._children[:-1]
        else:
            l = (self,)

        if isinstance(right, _BaseExpr) and right._op == op_type:
            r = right._children[:-1]
        else:
            r = (right,)

        return _create_operator_expression(l, r + (expr_end,), op_type)
    
    # unary operators

    def __abs__(self):
        return self._overload_op_unary(_ExprOp.ABS)

    def __floor__(self):
        return self._overload_op_unary(_ExprOp.FLOOR)

    def __ceil__(self):
        return self._overload_op_unary(_ExprOp.CEIL)

    # operators

    def __add__(self, right):
        return self._overload_op_va_args(right, _ExprOp.ADD)

    def __sub__(self, right: 'TypeAny'):
        return self._overload_op_va_args(right, _ExprOp.SUB)

    def __mul__(self, right: 'TypeAny'):
        return self._overload_op_va_args(right, _ExprOp.MUL)

    def __truediv__(self, right: 'TypeAny'):
        return self._overload_op_va_args(right, _ExprOp.DIV)

    def __floordiv__(self, right: 'TypeAny'):
        div_expr = self.__truediv__(right)
        return div_expr.__floor__()

    def __pow__(self, right: 'TypeAny'):
        return self._overload_op(right, _ExprOp.POW)

    def __mod__(self, right: 'TypeAny'):
        return self._overload_op(right, _ExprOp.MOD)

def _create_operator_expression(left_children: 'TypeChildren', right_children: 'TypeChildren', op_type: int):
    new_expr = _BaseExpr()
    new_expr._op = op_type
    new_expr._children = (*left_children, *right_children)
    return new_expr

class _GenericExpr(_BaseExpr):
    
    def __init__(self, op: _ExprOp, rt: 'TypeResultType', fixed: 'TypeFixed'):
        self._op = op
        self._rt = rt
        self._fixed = fixed
